--[[
	Copyright (C) 2017  Fuhrbolg

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>

	Author's note: Under no circumstances should this Addon be used under any situation.
	It is dangerous and will cause significant harm to your computer. You have been warned.
--]]

--[[
	MongoMon is an addon that has the following functionality:
	
		1) Updates UnitFrame portraits with a new portrait. All images are in .tga format
		and are located in the Res folder
		2) Plays a sound when player is awarded killing blow. Sound files are .mp3 format
		and are also located in the Res folder.
		3) Creates and displays a frame with a personalized scoreboard. This frame is draggable.
		If the player is ranked in the top 5 in killing blows a special icon will be displayed
		on the scoreboard.
		4) Nameplates will display Kill - Death score above players' names.
		5) Players ranked in the top 5 in killing blows will get a special icon above their
		nameplate as well.
		
	This addon uses the Ace3 framework. This is used to accomplish the following:
	
		1) Abstract persistent storage using AceDB.
		2) Generate the Options Interface page that is part of blizzard's UI.
		3) Interationalization support for multiple languages. All of these translation
		files are located in the Locale folder.
		4) Console support. This is trivial but meh, when in mexico.
		5) Im not sure if LibStub is part of the Ace3 framework but that is used to load
		libraries.
		6) LibNameplatesRegistry, again not sure if this part of Ace3. It is used to get
		access to player nameplates.
		7) LibButtonGlow, this library is used to make the scoreboard flash. Im way too
		fat and lazy to do that myself. The flash occurs after a killing blow.
		
	Known Issues:
	
		1) Responsiveness. The personalized scoreboard and nameplates update sporadically. 
		MongoMon hooks into COMBAT_LOG_EVENT_UNFILTERED and requests a battle ground score
		update when the sub event is UNIT_DIED. It also requests a battleground score
		when the player changes options and finally when out of combat which is
		the PLAYER_HEALTH_REGEN event.
		
	I photoshopped the majority of the images myself. They are all terrible. Except the
	one with the mad baby which is my nephew. Very cuddly. But also very, very mad
	at all times. The sounds are from Return to Castle Wolfenstein. Whats better than
	a nice german soldat praising you? Nothing. On that subject there is an excellent
	book written by Col. Hans Von Luck, who was a Panzer Commander in the Wehrmacht.
	True gentleman. And another book called "A Higher Call" about a Luftwaffe ace named
	Frans Stiegler. 
	
	Frans saw a banged up b17 flying low over the coast of France and instead of pwning
	it in his 109 he escorted it over German flak batteries so they wouldn't shoot it.
	The american b17 pilot charlie brown (heh) eventually made it back to england. In
	the early 2000's Charlie Brown posted an article in an aviation magazine talking
	about his strange encounter with the friendly Me109 pilot. Frans saw the article
	and they met in Florida.
	
	List of fonts: http://wow.gamepedia.com/API_FontInstance_SetFontObject
--]]

local ADDON_NAME, T = ...

--[[
 -	Library verification follows.
--]]
if not LibStub then
    error(ADDON_NAME, " requires LibStub")
    return
end

if not LibStub("AceAddon-3.0") then
    error(ADDON_NAME, " requires AceAddon-3.0")
    return
end

if not LibStub("AceGUI-3.0") then
    error(ADDON_NAME, " requires AceGUI-3.0")
    return
end

if not LibStub("AceLocale-3.0") then
    error(ADDON_NAME, " requires AceLocale-3.0")
    return
end

if not LibStub("CallbackHandler-1.0") then
    error(ADDON_NAME, " requires CallbackHandler-1.0")
    return
end

if not LibStub("LibNameplateRegistry-1.0") then
    error(ADDON_NAME, " requires LibNameplateRegistry-1.0")
    return
end

if not LibStub("LibButtonGlow-1.0") then
	error(ADDON_NAME, " requires LibButtonGlow-1.0")
	return
end

if not LibStub("AceConfig-3.0") then
	error(ADDON_NAME, " requires AceConfig-3.0")
	return
end

if not LibStub("AceDB-3.0") then
	error(ADDON_NAME, " requires AceDB-3.0")
	return
end

if not LibStub("AceConsole-3.0") then
	error(ADDON_NAME, " requires AceConsole-3.0")
	return
end

-- Localization
local L = LibStub("AceLocale-3.0"):GetLocale("MongoMon")

-- Our addon Table
MongoMon = {

	-- Our options database. Here we store all the flags for toggling sounds, nameplates, portraits, etc.
	db = {},

	-- All players in BG { "<PlayerName>-<ServerName>" = { "portrait" = "<PortraitIcon>", ... }, ...}
	players = {},
	
	-- Map of Nameplates, keyed on player name { "<PlayerName>-<ServerName>" = { "fontString" = "<FontString>", ... }, ...}
	plates = {},
	
	-- Map of player name, keyed on guid { "<GUID>" = "<PlayerName>", ... }
	guids = {},
	
	-- Ref to our particular player data structure
	player = nil,
	
	-- Boolean indicating we are in instanced PvP
	instancedPvp = false,
	
	-- Player's faction name
	PLAYER_FACTION = "",
	
	-- Player's name
	PLAYER_NAME = "",
	
	-- Player GUID
	PLAYER_GUID = "",
	
	-- Pet name fucking hunters
	PET_NAME = "",
	
	-- Only reward mongos whose rank is below this. We currently only have 5 portraits so.
	MAX_RANK = 5,
	
	-- Flag indicating whether we should toggle total damage or total healing on personal scoreboard.
	IS_HEALER = false,
	
	-- [GUID] = killTime (from GetTime())
	RECENT_KILLS = setmetatable({}, { __mode = "kv" }),
	
	-- Array of our sound files, played when target dies, praise be unto Hillary
	SOUND_FILES = {
		"Interface\\AddOns\\MongoMon\\Res\\Wunderbar.mp3",
		"Interface\\AddOns\\MongoMon\\Res\\GoodGame.mp3",
		"Interface\\AddOns\\MongoMon\\Res\\Sorry.mp3",
		"Interface\\AddOns\\MongoMon\\Res\\Excellent.mp3",
		"Interface\\AddOns\\MongoMon\\Res\\Jahahaha.mp3",
		"Interface\\AddOns\\MongoMon\\Res\\AufWiedersehen.mp3",
		"Interface\\AddOns\\MongoMon\\Res\\EnemyWeakened.mp3",
		"Interface\\AddOns\\MongoMon\\Res\\GreatShot.mp3",
		"Interface\\AddOns\\MongoMon\\Res\\JaWohl.mp3",
		"Interface\\AddOns\\MongoMon\\Res\\AllClear.mp3",
		"Interface\\AddOns\\MongoMon\\Res\\PathCleared.mp3",
		"Interface\\AddOns\\MongoMon\\Res\\Halo.mp3",
		"Interface\\AddOns\\MongoMon\\Res\\Nein.mp3",
		"Interface\\AddOns\\MongoMon\\Res\\JaWohl2.mp3"
	},
	
	-- Array of our terrible portraits
	IMAGE_FILES = {
		["Rogue"] = {
			["Assassination"] = "Interface\\AddOns\\MongoMon\\Res\\Broseph",
			["Outlaw"] = "Interface\\AddOns\\MongoMon\\Res\\Broseph",
			["Subtlety"] = "Interface\\AddOns\\MongoMon\\Res\\Broseph"
		},
		["Death Knight"] = {
			["Blood"] = "Interface\\AddOns\\MongoMon\\Res\\Emo",
			["Frost"] = "Interface\\AddOns\\MongoMon\\Res\\Emo",
			["Unholy"] = "Interface\\AddOns\\MongoMon\\Res\\Emo"
		},
		["Warrior"] = {
			["Fury"] = "Interface\\AddOns\\MongoMon\\Res\\MadBaby",
			["Arms"] = "Interface\\AddOns\\MongoMon\\Res\\MadBaby",
			["Protection"] = "Interface\\AddOns\\MongoMon\\Res\\MadBaby"
		},
		["Hunter"] = {
			["Survival"] = "Interface\\AddOns\\MongoMon\\Res\\DerpFace",
			["Beast Mastery"] = "Interface\\AddOns\\MongoMon\\Res\\DerpFace",
			["Marksmanship"] = "Interface\\AddOns\\MongoMon\\Res\\DerpFace"
		},
		["Priest"] = {
			["Shadow"] = "Interface\\AddOns\\MongoMon\\Res\\Kitten"
		},
		["Paladin"] = {
			["Retribution"] = "Interface\\AddOns\\MongoMon\\Res\\AngryGirl"
		},
		["Druid"] = {
			["Restoration"] = "Interface\\AddOns\\MongoMon\\Res\\CatNerd",
			["Feral"] = "Interface\\AddOns\\MongoMon\\Res\\CatLady",
			["Balance"] = "Interface\\AddOns\\MongoMon\\Res\\Hillary"
		},
		["Warlock"] = {
			["Affliction"] = "Interface\\AddOns\\MongoMon\\Res\\Impossibru",
			["Demonology"] = "Interface\\AddOns\\MongoMon\\Res\\Impossibru",
			["Destruction"] = "Interface\\AddOns\\MongoMon\\Res\\Impossibru"
		}
	}
}

-- Inject Ace3 Framework support
LibStub("AceAddon-3.0"):NewAddon(MongoMon, "MongoMon", "LibNameplateRegistry-1.0", "AceConsole-3.0")

-- Slash command /mm opens the options interface and prints out the version.
MongoMon:RegisterChatCommand("mm", "ConsoleCommand")

-- Matches any "unit" under the player's control, used by the COMBAT_LOG_UNFILTERED event handler to look for killing blows
-- borrowed from the KillingBlow_Enhanced addon. Muchos gracious.
local FILTER_MINE = bit.bor(COMBATLOG_OBJECT_AFFILIATION_MINE, COMBATLOG_OBJECT_REACTION_FRIENDLY, COMBATLOG_OBJECT_CONTROL_PLAYER)

-- Should always be disabled.
local DEBUG_ENABLED = false

--[[
 -	Options table follows. This defines the options that will be dispalyed on
 -	the Interface Options page within the Blizzard Addons menu.
--]]

-- Our AceDB default options
local defaults = {
	profile = {
		portraitEnabled = true,
		scoreEnabled = true,
		nameplatesEnabled = true,
		flashEnabled = true,
		soundEnabled = true
	}
}

-- Our AceDB options table
local options = {
	type = "group",
	name = "MongoMon",
	desc = L["Description"],
	handler = MongoMon,
	set = "OptionSetterHandler",
	get = "OptionGetterHandler",
	args = {
		optionsSubtitle = {
			order = 1,
			type = "header",
			name = L["Subtitle"]
		},
		optionsDescription = {
			order = 2,
			type = "description",
			name = L["Description"]
		},
		portraitTitle = {
			order = 3,
			type = "header",
			name = ""
		},
		portraitEnabled = {
			order = 4,
			type = "toggle",
			name = L["EnablePortrait"]
		},
		portraitDescription = {
			order = 5,
			type = "description",
			name = L["PortraitDescription"]
		},
		scoreTitle = {
			order = 6,
			type = "header",
			name = ""
		},
		scoreEnabled = {
			order = 7,
			type = "toggle",
			name = L["EnableScore"]
		},
		scoreDescription = {
			order = 8,
			type = "description",
			name = L["ScoreDescription"]
		},
		nameplatesTitle = {
			order = 9,
			type = "header",
			name = ""
		},
		nameplatesEnabled = {
			order = 10,
			type = "toggle",
			name = L["EnableNameplates"],
		},
		nameplatesDescription = {
			order = 11,
			type = "description",
			name = L["NameplatesDescription"]
		},
		flashTitle = {
			order = 12,
			type = "header",
			name = ""
		},
		flashEnabled = {
			order = 13,
			type = "toggle",
			name = L["EnableFlash"]
		},
		flashDescription = {
			order = 14,
			type = "description",
			name = L["FlashDescription"]
		},
		soundTitle = {
			order = 15,
			type = "header",
			name = ""
		},
		soundEnabled = {
			order = 16,
			type = "toggle",
			name = L["EnableSound"]
		},
		soundDescription = {
			order = 17,
			type = "description",
			name = L["SoundDescription"]
		}
	}
}

-- These are the two primary backgrounds for our scoreboard. The only difference is the border texture
-- which becomes more prominent when the player is in the top 5 killing blows. The gradient change 
-- occurs elsewhere and for different reasons.
local BACKDROP_RED = {
	bgFile = "Interface\\AddOns\\MongoMon\\Res\\RedAlphaBackground", 
	edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Border",
	edgeSize = 30,
	insets = { left = 8, right = 8, top = 8, bottom = 8 }
}

local BACKDROP_GREEN = {
	bgFile = "Interface\\AddOns\\MongoMon\\Res\\GreenAlphaBackground", 
	edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Border",
	edgeSize = 30,
	insets = { left = 8, right = 8, top = 8, bottom = 8 }
}

local BACKDROP_BAD = {
	bgFile = "Interface\\AddOns\\MongoMon\\Res\\BlackAlphaBackground", 
	edgeFile = "Interface\\DialogFrame\\UI-DialogBox-Border",
	edgeSize = 30,
	insets = { left = 8, right = 8, top = 8, bottom = 8 }
}

-- So everything in this game that is event related must be part of a widget.
local frame = CreateFrame("FRAME", "MongoMonFrame")

-- Scoreboard frame that will display Kill - Death in BG's
local score = CreateFrame("FRAME", "MongoMonScoreFrame")
score:SetMovable(true)
score:EnableMouse(true)
score:RegisterForDrag("LeftButton")
score:SetScript("OnDragStart", score.StartMoving)
score:SetScript("OnDragStop", score.StopMovingOrSizing)
score:SetPoint("CENTER")
score:SetWidth(200)
score:SetHeight(70)
score:SetBackdrop(BACKDROP_BAD)

-- Top Damage icon container on score board
score.topDamageFrame = CreateFrame("FRAME", nil, score)
score.topDamageFrame:SetPoint("CENTER")
score.topDamageFrame:SetSize(125, 55)
score.topDamageFrame:Hide()

-- Top Damage icon iteself. This is an animated sprite. See below.
score.topDamageTexture = score.topDamageFrame:CreateTexture()
score.topDamageTexture:SetAlpha(0.25)
score.topDamageTexture:SetTexture("Interface\\AddOns\\MongoMon\\Res\\Wheelchair")
score.topDamageTexture:SetAllPoints()
score.topDamageTexture:SetTexCoord(0, 0.25, 0, 0.25)

-- Title 
score.title = score:CreateFontString(nil, "ARTWORK", "GameFontHighlightOutline")
score.title:SetPoint("BOTTOM", score, "TOP", 0, -5) -- Places frame in center of screen
score.title:SetText(L["Title"])

-- Fontstring
score.fontString = score:CreateFontString(nil, "ARTWORK", "GameFontNormalHugeOutline2")
score.fontString:SetPoint("TOP", score, 0, -15) -- Places frame in center of screen

-- Damage fontstring
score.damageFontString = score:CreateFontString(nil, "ARTWORK", "GameFontHighlightSmallOutline")
score.damageFontString:SetPoint("BOTTOM", score, 0, 15) -- Places frame in center of screen

-- High Rank Crest Texture
score.crestFrame = CreateFrame("FRAME", nil, score)
score.crestFrame:Hide()
score.crestFrame:SetFrameLevel(127) -- For issue #1, the frame layer was spastic. Occasionally displayed behind the rank crest texture.
score.crestFrame:SetWidth(60)
score.crestFrame:SetHeight(60)
score.crestFrame:SetPoint("BOTTOMLEFT", score, "TOPLEFT", -3, -10)
score.crestTex = score.crestFrame:CreateTexture()
score.crestTex:SetTexture("Interface\\AddOns\\MongoMon\\Res\\RankCrest")
score.crestTex:SetAllPoints()

-- High Rank Texture
score.rankFrame = CreateFrame("FRAME", nil, score.crestFrame)
score.rankFrame:SetFrameLevel(128) -- For issue #1, the frame layer was spastic. Occasionally displayed behind the rank crest texture.
score.rankFrame:SetWidth(40)
score.rankFrame:SetHeight(40)
score.rankFrame:SetPoint("CENTER", score.crestFrame)
score.rankTex = score.rankFrame:CreateTexture()
score.rankTex:SetAllPoints()

--[[
 -	Utility function used for sorting.
 -
 - An ordered sort. This is exceptionally handy and is used to order
 - the battleground participants by killing blows then defaults to 
 - lower number of deaths in the event of a tie. This particular
 - function is a generic sort. The man who wrote this must be jesus
 - because I have no idea what's going on. It just works. Praise Hillary.
 -
 - http://stackoverflow.com/questions/15706270/sort-a-table-in-lua
--]]
local function spairs(t, order)
    -- collect the keys
    local keys = {}
    for k in pairs(t) do keys[#keys+1] = k end

    -- if order function given, sort by it by passing the table and keys a, b,
    -- otherwise just sort the keys 
    if order then
        table.sort(keys, function(a,b) return order(t, a, b) end)
    else
        table.sort(keys)
    end

    -- return the iterator function
    local i = 0
    return function()
        i = i + 1
        if keys[i] then
            return keys[i], t[keys[i]]
        end
    end
end

--[[
 - Converts long numbers into an abreviated format, e.g. 46,387,123 into 46M
 - Hijacked from the internet
--]]
local function prettyNumber(num)
	if num <= 9999 then
		return v
	elseif num >= 1000000 then
		return format("%.1f mil", num/1000000)
	elseif num >= 10000 then
		return format("%.1fk", num/1000)
	end
	return num
end

--[[
 - Callback for OnUpdate to animate the sprite TGA image visual effect.
 - This is used by the score board and nameplates.
--]]
local function topDamageAnimation(self, elapsed)
	AnimateTexCoords(self.topDamageTexture, 512, 512, 256, 256, 4, elapsed, 0.10)
end

--[[
 -	This is an event handler for the Ace3 framework. When the addon is first loaded
 -	it will fire. We are just going to use it to handle the loading of our 
 -	options database.
--]]
function MongoMon:OnInitialize() 
	-- Generate the database of stored values
	self.db = LibStub("AceDB-3.0"):New("MongoMonDB", defaults, true).profile
	-- Register the options. This library handles all of the magic.
	LibStub("AceConfig-3.0"):RegisterOptionsTable("MongoMon", options)
	-- Generate the Options Interface
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions("MongoMon", "MongoMon")
end

--[[
 -	Callback when an option is set in the Interface Options menu.
 -	
 -	Use this to update our internal storage.
--]]
function MongoMon:OptionSetterHandler(info, value)
	self.db[info[#info]] = value
	if self.instancedPvp then -- Re-initialize state if user changes settings during a battleground
		self:Initialize()
	end
end

--[[
 -	Callback when to get the value of an options item.
--]]
function MongoMon:OptionGetterHandler(info)
	return self.db[info[#info]]
end

--[[
 -	Prepares the Addon by registering the required events
 -	when a player enters a zone or logs into the game world.
 -	
 -	Has the reverse effect when player zones out of PvP by unregistering
 -	events.
 -	
 -	Note: We can include the UPDATE_WORLD_STATE event to queue the
 -	server for BG data if need be.
 -	
 -	TODO: Disable for arenas?
 -	isArena, isRegistered = IsActiveBattlefieldArena()
--]]
function MongoMon:Initialize()

	-- Set the instanced PvP flag
	local _, instanceType = IsInInstance()
	self.instancedPvp = instanceType == "pvp"
	
	-- Clears all refs
	table.wipe(self.players)
	table.wipe(self.guids)
	self.player = nil
	
	self.PLAYER_NAME = UnitName("player")
	self.PLAYER_GUID = UnitGUID("player")
	self.PET_NAME = UnitName("pet")
	self.PLAYER_FACTION = UnitFactionGroup("player")

	if self.instancedPvp then 
		-- Register events
		frame:RegisterEvent("UPDATE_BATTLEFIELD_SCORE") 	-- Portraits, Nameplates
		frame:RegisterEvent("PLAYER_REGEN_ENABLED")		-- Portraits, Nameplates
		frame:RegisterEvent("PLAYER_TARGET_CHANGED")		-- Portraits
		frame:RegisterEvent("UNIT_PORTRAIT_UPDATE")		-- Portraits
		frame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")	-- Sounds, Guids, Killing Blows
		
		-- Lib Nameplate Registry callbacks
		self:LNR_RegisterCallback("LNR_ON_NEW_PLATE")			-- Nameplates
		self:LNR_RegisterCallback("LNR_ON_RECYCLE_PLATE")		-- Nameplates
		self:LNR_RegisterCallback("LNR_ON_TARGET_PLATE_ON_SCREEN")	-- Nameplates
		self:LNR_RegisterCallback("LNR_ERROR_FATAL_INCOMPATIBILITY")	-- Nameplates
		
		self.IS_HEALER = MongoMon:IsPlayerHealer()
		
		if self.db.scoreEnabled then
			score.fontString:SetText("0 - 0")
			score.damageFontString:SetText("")
			score:Show()
		end
	elseif DEBUG_ENABLED then
		score:SetScript("OnMouseUp", function(self, button)
			LibStub("LibButtonGlow-1.0").ShowOverlayGlow(score)
			LibStub("LibButtonGlow-1.0").HideOverlayGlow(score)	
			score.damageFontString:SetFontObject("GameFontNormal")
		end)
		score.crestFrame:Show()
		score.fontString:SetText("27 - 52 Rank #5")
		score.damageFontString:SetText("48M damage, Rank #2")
		score.rankTex:SetTexture("Interface\\AddOns\\MongoMon\\Res\\Rank1")
		score.topDamageFrame:Show()
		score:SetScript("OnUpdate", topDamageAnimation)
		score:SetBackdrop(BACKDROP_GREEN)
		score:Show()
		
	else 
		-- Events are no longer needed lets be nice and clean them up
		frame:UnregisterEvent("UPDATE_BATTLEFIELD_SCORE")
		frame:UnregisterEvent("PLAYER_REGEN_ENABLED")
		frame:UnregisterEvent("PLAYER_TARGET_CHANGED")
		frame:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
		self:LNR_UnregisterAllCallbacks()
		score:Hide()
	end
end

--[[
 -	Queries to check if the player is currently a healer so we can
 -	provide healing stats instead of damage stats in a BG.
--]]
function MongoMon:IsPlayerHealer()
	local currentSpec = GetSpecialization()
	local currentSpecName = currentSpec and select(2, GetSpecializationInfo(currentSpec)) or nil
	local items = { "Discipline", "Holy", "Restoration", "Mistweaver" }
	
	for _,v in pairs(items) do
		if v == currentSpecName then
			return true
		end
	end
	return false
end

--[[
 - Completely Hijacked from Choonster's KillingBlow_Enhanced addon. This is much better
 - than my original attempts at computing killing blows. This will play a sound and flash the BG
 - scoreboard window.
--]]
function MongoMon:KillingBlowCheck(timestamp, event, hideCaster, sourceGUID, sourceName, sourceFlags, sourceRaidFlags, destGUID, destName, destFlags, destRaidFlags, ...)
	if
		not destGUID or destGUID == "" or -- If there isn't a valid destination GUID
		(sourceGUID ~= self.PLAYER_GUID and bit.band(sourceFlags, FILTER_MINE) ~= FILTER_MINE) or -- Or the source unit isn't the player or something controlled by the player (the latter check was suggested by Caellian)
		not destGUID:find("^Player%-") -- Or we're in a PvP zone and the destination unit isn't a player
	then return end -- Return now

	local _, overkill
	if event == "SWING_DAMAGE" then
		_, overkill = ...
	elseif event:find("_DAMAGE", 1, true) and not event:find("_DURABILITY_DAMAGE", 1, true) then
		_, _, _, _, overkill = ...
	end

	local now, previousKill = GetTime(), self.RECENT_KILLS[destGUID]

	-- Caellian has noted that PARTY_KILL doesn't always fire correctly and suggested checking the overkill argument
	-- (which will be 0 [or maybe -1] for non-killing blows) to mitigate against this.
	--
	-- Because most kills will trigger PARTY_KILL and an overkill _DAMAGE, we need to keep a record of recent kill times
	-- and only record kills of the same unit when they're at least 2 seconds apart.
	if (event == "PARTY_KILL" or (overkill and overkill > 0)) and (not previousKill or now - previousKill > 2.0) then
		local index = math.random(1, table.getn(self.SOUND_FILES))
		local soundFile = self.SOUND_FILES[index]
		PlaySoundFile(soundFile, "master")
		self.RECENT_KILLS[destGUID] = now
	end
end

--[[
 -	Handles PLAYER_TARGET_CHANGED events.
--]]
function MongoMon:PlayerTargetChanged()

	if not self.db.portraitEnabled then return end	
	if not self.instancedPvp then return end

	if not UnitIsFriend("player", "target") then -- Is target hostile
		if not UnitIsDead("target") then -- Is unit alive

			local name = self:getTargetName()
			
			if not name then return end
			
			local portrait = (self.players[name] and self.players[name].portrait) and self.players[name].portrait or nil
			
			if portrait then -- If nil then it is not overpowered
				TargetFrame.portrait:SetTexture(portrait)
			end
		end
	end	
end

--[[
 -	This is a response from the server to our request battleground score
 -	API call. We are going to cache this data in the MongoMon table.
 -
 -	Here we will set or update our players array and do some calculation to
 -	that determines ranking in the BG and whether the players qualify for
 -	a special nameplate.
--]]
function MongoMon:UpdateBattlefieldScore()

	-- Count of BG participants
	local numBgPlayers = GetNumBattlefieldScores()
	local originalPlayerKills = 0
	if self.player then
		originalPlayerKills = self.player.kills
	end
	for x = 1, numBgPlayers, 1 do 
		name, killingBlows, honorableKills, deaths, honorGained, faction, race, class, classToken, damageDone, healingDone, bgRating, ratingChange, preMatchMMR, mmrChange, talentSpec = GetBattlefieldScore(x)

		-- Around 6.0 blizzard started sending nil combatants into bgs, this is very rng based.
		-- Look around, you'll see the scoreboard has empty player names. GG.
		if name then 
			local player = self.players[name]
			
			if player then
				player.kills = killingBlows
				player.deaths = deaths
				player.damage = damageDone
				player.healing = healingDone
			else
				faction = faction == 0 and "Horde" or "Alliance"
				player = {
					name = name,
					kills = killingBlows,
					deaths = deaths,
					damage = damageDone,
					healing = healingDone,
					class = class,
					spec = talentSpec,
					portrait = (self.IMAGE_FILES and self.IMAGE_FILES[class] and self.IMAGE_FILES[class][talentSpec]) and self.IMAGE_FILES[class][talentSpec] or nil,
					enemy = faction ~= self.PLAYER_FACTION,
					rank = 0,
					damageRank = 0,
					healingRank = 0,
					GUID = self:GetGUIDByName(name)
				}				
			end
			
			-- Keyed as such: name-server
			self.players[name] = player 
			
			-- One time assignment of our player object to the global reference
			if not self.player and player.name == self.PLAYER_NAME then
				self.player = player
			end
		end
	end
	
	-- Sorts BG scores by top kills and iterates in descending order
	local topRanks = 0 
	for _, player in spairs(self.players, function(t,a,b) 
			if t[b].kills == t[a].kills then return t[b].deaths > t[a].deaths -- Tie goes to player with less deaths
			else return t[b].kills < t[a].kills end
	end) do
		topRanks = topRanks + 1
		player.rank = topRanks
	end
	
	-- Sorts BG scores by healing done
	local healingRank = 0
	for _, player in spairs(self.players, function(t,a,b) 
			if t[b].healing == t[a].healing then return t[b].deaths > t[a].deaths -- Tie goes to player with less deaths
			else return t[b].healing < t[a].healing end
		end) 
	do
		healingRank = healingRank + 1
		player.healingRank = healingRank
	end
	
	-- Sorts BG scores by damage
	local damageRank = 0
	for _, player in spairs(self.players, function(t,a,b) 
			if t[b].damage == t[a].damage then return t[b].deaths > t[a].deaths -- Tie goes to player with less deaths
			else return t[b].damage < t[a].damage end
		end) 
	do
		damageRank = damageRank + 1
		player.damageRank = damageRank
		
	end
	
	-- Make all UI Updates to Nameplates and to player scoreboard
	for _, player in pairs(self.players) do

		-- Update nameplates for all players
		self:UpdateNameplate(player)
		
		-- Update personal scoreboard
		if player.name == self.PLAYER_NAME then
			if self.IS_HEALER then
				-- Update scoreboard for healing
				if player.healing == nil or player.healing == 0 then
					score.damageFontString:SetText("0 healing")
				elseif player.healingRank > 3 then -- Standard font for people not in the top tier
					score.damageFontString:SetFontObject("GameFontHighlightSmallOutline")
					score.damageFontString:SetText(prettyNumber(player.healing) .. " healing, Rank #" .. player.healingRank)
				else
					score.damageFontString:SetFontObject("GameFontNormal")
					score.damageFontString:SetText(prettyNumber(player.healing) .. " healing, Rank #" .. player.healingRank)
				end			
			
				-- Special case for #1 Healing. Update scoreboard animation of power!
				if player.healingRank == 1 and player.healing > 0 then -- Number 1 healing
					score:SetBackdrop(BACKDROP_GREEN)
					score.topDamageFrame:Show()
					score:SetScript("OnUpdate", topDamageAnimation)
				else -- Not number 1 healing
					score:SetBackdrop(BACKDROP_BAD)
					score:SetScript("OnUpdate", nil)
					score.topDamageFrame:Hide()
				end
			else
				-- Update scoreboard for damage
				if player.damage == nil or player.damage == 0 then
					score.damageFontString:SetText("0 damage")
				elseif player.damageRank > 3 then -- Standard font for people not in the top tier
					score.damageFontString:SetFontObject("GameFontHighlightSmallOutline")
					score.damageFontString:SetText(prettyNumber(player.damage).. " damage, Rank #" .. player.damageRank)
				else
					score.damageFontString:SetFontObject("GameFontNormal")
					score.damageFontString:SetText(prettyNumber(player.damage).. " damage, Rank #" .. player.damageRank)
				end					
			
				-- Special case for #1 Damage. Update scoreboard animation of power!
				if player.damageRank == 1 and player.damage > 0 then -- Number 1 damage
					score:SetBackdrop(BACKDROP_RED)
					score.topDamageFrame:Show()
					score:SetScript("OnUpdate", topDamageAnimation)
				else -- Not number 1 damage
					score:SetBackdrop(BACKDROP_BAD)
					score:SetScript("OnUpdate", nil)
					score.topDamageFrame:Hide()
				end
			end
			
			-- This section will update the player's personal scoreboard for killing blows and deaths
			-- as well as update their personal rank icon.
			local rank = player.kills > 0 and "Rank #" .. player.rank or ""
			score.fontString:SetText(player.kills .. " - " .. player.deaths .. " " .. rank)
			
			-- Highlight score frame
			if player.kills > originalPlayerKills and self.db.flashEnabled then
				LibStub("LibButtonGlow-1.0").ShowOverlayGlow(score)
				LibStub("LibButtonGlow-1.0").HideOverlayGlow(score)
			end
			
			if player.rank > 0 and player.rank <= self.MAX_RANK and player.kills > 0 then
				score.rankTex:SetTexture("Interface\\AddOns\\MongoMon\\Res\\Rank" .. player.rank)
				score.crestFrame:Show()
				score.rankFrame:Show()
			else
				score.crestFrame:Hide()
				score.rankFrame:Hide()
			end
		end		
	end
end

--[[ 
 -	Helper function to get the compelte target name.
 -	
 -	Gets the complete name of the currently selected target
 -	Returns: Name - Server
--]]
function MongoMon:getTargetName()

	local targetName, targetRealm = UnitName("target")
		
	-- So this is kind of crappy, players of the same realm will have a nil realm. 
	-- Check for realm and append to name for the love of stalin.
	if targetRealm then
		targetName = targetName .. "-" .. targetRealm
	end	
	
	return targetName
end

--[[
 - Jesus christ this is a piece of shit. We need to listen to all the trash in the combat
 - log to get a handle on a player guid because fucking cocksucking GetBattlefieldScores 
 - doesn't supply it. Fucking piece of shit.
--]]
function MongoMon:CaptureGuids(sourceName, sourceGUID, sourceFlags, destName, destGUID, destFlags) 
	if not self.instancedPvp then return end
	
	if sourceGUID and sourceName and sourceGUID:find("^Player%-") and sourceFlags and bit.band(sourceFlags, 66888) then
		if not self.guids[sourceGUID] then
			self.guids[sourceGUID] = sourceName
		end
	end	
	
	if destGUID and destName and destFlags and bit.band(destFlags, 66888) then
		if destGUID ~= sourceGUID and destName ~= sourceName and destGUID:find("^Player%-") then
			if not self.guids[destGUID] then
				self.guids[destGUID] = destName
			end
		end
	end	
end

--[[
 -	Gets a player GUID by the player name. Player name is server-name.
 -
 -	GUID is captured by sniffing through the combat log.
--]]
function MongoMon:GetGUIDByName(name)
	for playerGUID, playerName in pairs (self.guids) do
		if playerName == name then
			return playerGUID
		end
	end
	return nil
end

--[[
 -	Updates an existing plate if it is visible.
 -	Will check for Killing Blow rank, score and Damage rank
 -
--]]
function MongoMon:UpdateNameplate(player)

	if not player then return end
	
	local plateFrame = self:GetPlateByGUID(player.GUID)
	
	if not plateFrame or not plateFrame.MM or not plateFrame.MM.topDamageFrame or not plateFrame.MM.rankTexture or not plateFrame.MM.scoreFontString then
		return
	end
	
	-- Player Kill / Death score
	plateFrame.MM.scoreFontString:Show()
	plateFrame.MM.scoreFontString:SetText(player.kills .. " - " .. player.deaths )
	
	-- Killing Blow Icon
	if player.rank > 0 and player.rank <= self.MAX_RANK and player.kills > 0 then
		plateFrame.MM.rankTexture:SetTexture("Interface\\AddOns\\MongoMon\\Res\\Rank" .. player.rank)
		plateFrame.MM.rankTexture:Show()
	else
		plateFrame.MM.rankTexture:Hide()
	end
	
	-- Top Damage Icon
	if player.damage > 0 and player.damageRank == 1 then
		plateFrame.MM.topDamageFrame:Show()
		plateFrame.MM.topDamageFrame.topDamageTexture:Show()
		plateFrame.MM.topDamageFrame:SetScript("OnUpdate", topDamageAnimation)	
	else
		plateFrame.MM.topDamageFrame:SetScript("OnUpdate", nil)
		plateFrame.MM.topDamageFrame.topDamageTexture:Hide()
		plateFrame.MM.topDamageFrame:Hide()
	end
end

--[[
 -	Event handler for LibNameplateRegistry's LNR_ON_NEW_PLATE event.
 -	
 -	A new plate is in range. Let us do things.
 -
 -	https://www.wowace.com/addons/libnameplateregistry-1-0/pages/api/
--]]
function MongoMon:LNR_ON_NEW_PLATE(eventname, plateFrame, plateData)

	if not plateFrame or not plateData or not plateData.GUID or not plateData.name or plateData.type ~= "PLAYER" or not self.instancedPvp then return end

	local name = self.guids[plateData.GUID]
	
	if not name then return end

	local player = self.players[name]
	
	if not player then -- Player not found
		return 
	elseif self.PLAYER_GUID == plateData.GUID then
		return	
	elseif  self.db.nameplatesEnabled then
	
		-- Create Fontstring for Kill - Death score
		local scoreFontString = plateFrame:CreateFontString()
		scoreFontString:SetFont(_G.NumberFont_Shadow_Small:GetFont(), 15, "THICKOUTLINE")
		if player.enemy then
			scoreFontString:SetTextColor(1, 0.1, 0.1, 0.75)
		else
			scoreFontString:SetTextColor(0, 1, 0, 1)
		end
		scoreFontString:SetPoint("CENTER", 0, 35)
		scoreFontString:SetText(player.kills .. " - " .. player.deaths)
		
		local rankTexture = plateFrame:CreateTexture()
		rankTexture:SetWidth(32)
		rankTexture:SetHeight(32)
		rankTexture:SetPoint("TOPLEFT", plateFrame, "TOPLEFT", 0, 25)
		rankTexture:Hide()

		if player.rank > 0 and player.rank <= self.MAX_RANK and player.kills > 0 then
			rankTexture:SetTexture("Interface\\AddOns\\MongoMon\\Res\\Rank" .. player.rank)
			rankTexture:Show()
		end

		-- Top Damage Texture	
		local topDamageFrame = CreateFrame("FRAME", nil, plateFrame)
		topDamageFrame:SetPoint("TOPRIGHT", plateFrame, "TOPRIGHT", -10, 25)
		topDamageFrame:SetWidth(34)
		topDamageFrame:SetHeight(34)
		
		-- Top Damage Texture
		topDamageFrame.topDamageTexture = topDamageFrame:CreateTexture()
		topDamageFrame.topDamageTexture:SetPoint("CENTER")
		topDamageFrame.topDamageTexture:SetAllPoints()
		topDamageFrame.topDamageTexture:SetTexture("Interface\\AddOns\\MongoMon\\Res\\Wheelchair")
		topDamageFrame.topDamageTexture:SetTexCoord(0, 0.25, 0, 0.25)
		topDamageFrame:Hide()
		topDamageFrame.topDamageTexture:Hide()
		
		if player.damage > 0 and player.damageRank == 1 then
			topDamageFrame:Show()
			topDamageFrame.topDamageTexture:Show()
			topDamageFrame:SetScript("OnUpdate", topDamageAnimation)
		end	

		plateFrame.MM = {
			scoreFontString = scoreFontString,
			rankTexture = rankTexture,
			topDamageFrame = topDamageFrame
		}
	end
end

--[[
 -	Event handler for LibNameplateRegistry's LNR_ON_RECYCLE_PLATE event.
 -	
 -	I believe this means the player has gone out of range. Not sure if
 -	we care about that.
--]]
function MongoMon:LNR_ON_RECYCLE_PLATE(eventname, plateFrame, plateData)
	if plateFrame and plateFrame.MM then
		plateFrame.MM.scoreFontString:Hide()
		plateFrame.MM.rankTexture:Hide()
		plateFrame.MM.topDamageFrame:Hide()
		plateFrame.MM.topDamageFrame.topDamageTexture:Hide()
		plateFrame.MM.topDamageFrame:SetScript("OnUpdate", nil)
		plateFrame.MM = nil
	end
end

--[[
 - 	Not entirely sure about this one here.
 -
 -	Perhaps one day we will need it.
--]]
function MongoMon:LNR_ON_TARGET_PLATE_ON_SCREEN(eventname, plateFrame, plateData)

end

--[[
 -	Event handler for LibNameplateRegistry's LNR_ERROR_FATAL_INCOMPATIBILITY event.
 -	
 -	Make sure we are using a compatible version of LNR.
--]]
function MongoMon:LNR_ERROR_FATAL_INCOMPATIBILITY(eventname, icompatibilityType)
	print("There is a fatal error: " .. eventname .. " Error: " .. icompatibilityType)
end

--[[
 -	Event handler for LibNameplateRegistry's LNR_ERROR_FATAL_INCOMPATIBILITY event.
--]]
function MongoMon:LNR_ERROR_FATAL_INCOMPATIBILITY(eventname, icompatibilityType)
	-- Here you want to check if your add-on and LibNameplateRegistry are not
	-- outdated (old TOC). if they're both up to date then it means that
	-- another add-on author thinks his add-on is more important than yours. In
	-- this later case you can register LNR_ERROR_SETPARENT_ALERT and
	-- LNR_ERROR_SETSCRIPT_ALERT which will detect such behaviour and will give
	-- you the name of the incompatible add-on so you can inform your users properly
	-- about what's happening instead of just silently "not working".
	print(eventname, icompatibilityType)
end

--[[
 - Command line response handler for the /mm command.
 -
 - Displays version and opens Addon window options.
--]]
function MongoMon:ConsoleCommand(input)
	InterfaceOptionsFrame_OpenToCategory(ADDON_NAME)
	InterfaceOptionsFrame_OpenToCategory(ADDON_NAME)
	print(L["Console"] .. " Version " .. GetAddOnMetadata("MongoMon", "Version") .. ".")
end

--[[
 -	Main Addon event listener.
 -
 -	A few notes on RequestBattlefieldScoreData:
 -	
 -	Polls the WoW server for Battleground data. The response is
 -	asynchronous in the form of an UPDATE_BATTLEFIELD_SCORE event.
 -	
 -	The same behavior can be achieved by hitting shift + spacebar to manually
 -	inspect the bg data. This was making me so mad.
 -	
 -	So now I understand why kunda (BattleGroundTargets) does what he does
 -	with the PLAYER_REGEN_ENABLED event. This signifies the player is out of
 -	combat and happens at a regular enough frequency to capture the bg score data.	
--]]
local function EventHandler(self, event, ...)
	if event == "COMBAT_LOG_EVENT_UNFILTERED" then
		local timestamp, type, hideCaster, sourceGUID, sourceName, sourceFlags, sourceFlags2, destGUID, destName, destFlags, destFlags2 = select(1, ...)
		MongoMon:CaptureGuids(sourceName, sourceGUID, sourceFlags, destName, destGUID, destFlags)	
		if type == "UNIT_DIED" or type == "PARTY_KILL" then
			RequestBattlefieldScoreData()		
		end
		MongoMon:KillingBlowCheck(...)	
	elseif event == "PLAYER_TARGET_CHANGED" then
		MongoMon:PlayerTargetChanged()
	elseif event == "UNIT_PORTRAIT_UPDATE" then
		MongoMon:PlayerTargetChanged()
	elseif event == "PLAYER_REGEN_ENABLED" then
		RequestBattlefieldScoreData()
	elseif event == "UPDATE_BATTLEFIELD_SCORE" then
		MongoMon:UpdateBattlefieldScore()
	elseif event == "PLAYER_ENTERING_WORLD" then
		MongoMon:Initialize()
		RequestBattlefieldScoreData()
	end
end

-- Always listening
frame:RegisterEvent("PLAYER_ENTERING_WORLD")
frame:SetScript("OnEvent", EventHandler)